module L = Llvm
module A = Ast
open Sast
module StringMap = Map.Make(String)
let translate(functions) =
  let context = L.global_context() in
  let the_module = L.create_module context "FUNC-y Java" in
    let i32_t      = L.i32_type    context
    and i8_t       = L.i8_type     context
    and i1_t       = L.i1_type     context
    and double_t    = L.double_type context in
    let str_t      = L.pointer_type i8_t
    and cry_t = L.void_type context in

  let ltype_of_typ = function
    A.Int    -> i32_t
    | A.Bool   -> i1_t
    | A.Float  -> double_t
    | A.Cry    -> cry_t
    | A.String -> str_t
    | A.Char -> i8_t
    | A.IntList (t) -> L.array_type i32_t t
    | A.FloatList (t) -> L.array_type double_t t
    | A.StringList (t) -> L.array_type str_t t

  in
  let type_of_lltype typ =
    let ltype_str = L.string_of_lltype typ in
    match ltype_str  with
     "i32"     -> A.Int
     | "i8"    -> A.Char
     | "i1"    -> A.Bool
     | "double" -> A.Float
     | "i8*"  -> A.String
     | "cry"   -> A.Cry
     | _ -> raise (Failure ("error: this type is unsupported"))
  in
  let type_of_lval lval =
    let lltype = L.type_of lval in
    type_of_lltype lltype
  in
let printf_t = L.var_arg_function_type i32_t [| L.pointer_type i8_t |] in
let printf_func = L.declare_function "printf" printf_t the_module in

  let function_decls : (L.llvalue * sfunc_decl) StringMap.t =
    let function_decl m fdecl =
      let name = fdecl.sfname
      and formal_types =
  Array.of_list (List.map (fun (t,_) -> ltype_of_typ t) fdecl.sformals)
      in let ftype = L.function_type (ltype_of_typ fdecl.styp) formal_types in
      StringMap.add name (L.define_function name ftype the_module, fdecl) m in
    List.fold_left function_decl StringMap.empty functions  in
  let build_function_body m_locals fdecl =
    let (the_function, _) = StringMap.find fdecl.sfname function_decls in
    let builder = L.builder_at_end context (L.entry_block the_function) in
    let str_format_str = L.build_global_stringptr "%s\n" "fmt" builder
    and int_format_str = L.build_global_stringptr "%d\n" "fmt" builder
    and float_format_str = L.build_global_stringptr "%f\n" "fmt" builder
    (* and bool_format_str = L.build_global_stringptr "%d\n" "fmt" builder *)
    and true_str = L.build_global_stringptr "true\n" "fmt" builder
    and false_str = L.build_global_stringptr "false\n" "fmt" builder
    and char_format_str = L.build_global_stringptr "%c\n" "fmt" builder in
    let add_formal m (t, n) p =
      L.set_value_name n p;
    	let local = L.build_alloca (ltype_of_typ t) n builder in
      ignore (L.build_store p local builder);
    	StringMap.add n local m
    in
    let locals = List.fold_left2 add_formal m_locals fdecl.sformals
        (Array.to_list (L.params the_function)) in
    (* Return the value for a variable or formal argument by
       checking local names *)
    let lookup n m =
      try StringMap.find n m
        with Not_found -> raise(Failure ("Variable " ^ n ^ " not declared") )
    in
    let extractValues n = match n with
      SNum(v) -> v
      | _ -> 0
    in
    let extractFValues n = match n with
      SDec(v) -> v
      | _ -> 0.0
    in
    let extractSValues n = match n with
      SSlit(v) -> v
      | _ -> ""
    in
    let rec expr builder m_locals ((_, e): sexpr) = match e with
      SNum i -> L.const_int i32_t i
      | SDec d -> L.const_float double_t d
      | SBlit b -> L.const_int i1_t (if b then 1 else 0)
      | SSlit s -> L.build_global_stringptr s "strpr" builder
      | SChara c -> L.const_int i8_t (Char.code c)
      | SArrayAccess(name, index) ->
        let list_index = (expr builder m_locals index)
        in let value = L.build_gep (lookup name m_locals) [| (L.const_int i32_t 0); list_index |] "tmp" builder
        in L.build_load value "tmp" builder
      | SVar v -> L.build_load (lookup v m_locals) v builder
      | SAssign (s, e) -> let e' = expr builder m_locals e in
                          ignore(L.build_store e' (lookup s m_locals) builder); e'
            | SBinop ((A.Float,_ ) as e1, op, e2) ->
        let e1' = expr builder m_locals e1
        and e2' = expr builder m_locals e2 in
        (match op with
            A.Add     -> L.build_fadd
          | A.Sub     -> L.build_fsub
          | A.Mult    -> L.build_fmul
          | A.Div     -> L.build_fdiv
          | A.Equal   -> L.build_fcmp L.Fcmp.Oeq
          | A.Neq     -> L.build_fcmp L.Fcmp.One
          | A.Less    -> L.build_fcmp L.Fcmp.Olt
          | A.Leq     -> L.build_fcmp L.Fcmp.Ole
          | A.Greater -> L.build_fcmp L.Fcmp.Ogt
          | A.Geq     -> L.build_fcmp L.Fcmp.Oge
          | A.And | A.Or -> raise (Failure "internal error: semant should have rejected and/or on float")
          | _ -> raise (Failure ("error: this operator is unsupported"))
        ) e1' e2' "tmp" builder
      | SBinop (e1, op, e2) ->
        let e1' = expr builder m_locals e1
        and e2' = expr builder m_locals e2 in
        (match op with
            A.Add     -> L.build_add
          | A.Sub     -> L.build_sub
          | A.Mult    -> L.build_mul
          | A.Div     -> L.build_sdiv
          | A.Mod     -> L.build_srem
          | A.And     -> L.build_and
          | A.Or      -> L.build_or
          | A.Equal   -> L.build_icmp L.Icmp.Eq
          | A.Neq     -> L.build_icmp L.Icmp.Ne
          | A.Less    -> L.build_icmp L.Icmp.Slt
          | A.Leq     -> L.build_icmp L.Icmp.Sle
          | A.Greater -> L.build_icmp L.Icmp.Sgt
          | A.Geq     -> L.build_icmp L.Icmp.Sge
        ) e1' e2' "tmp" builder
      | SUnop(op, ((t, _) as e)) ->
        let e' = expr builder m_locals e in
        (match op with
            A.Neg when t = A.Float -> L.build_fneg
          | A.Neg                  -> L.build_neg
          | A.Not                  -> L.build_not) e' "tmp" builder
      | SIntListLit(literals) ->
        let arrayValues = List.map(fun el -> snd(el)) literals in
        let v = List.map extractValues arrayValues in
        let values = List.map(L.const_int i32_t) v
        in let reversed_v = List.rev values in
        L.const_array i32_t (Array.of_list reversed_v)
      | SFloatListLit(literals) ->
        let arrayValues = List.map(fun el -> snd(el)) literals in
        let v = List.map extractFValues arrayValues in
        let values = List.map(L.const_float double_t) v
        in let reversed_v = List.rev values in
        L.const_array double_t (Array.of_list reversed_v)
      | SStringListLit(literals) ->
        let arrayValues = List.map(fun el -> snd(el)) literals in
        let v = List.map extractSValues arrayValues in
         let sexprList = List.map (fun e -> (A.String, SSlit(e))) v in
         let values = List.map (fun f -> expr builder m_locals f) sexprList
        in let reversed_v = List.rev values in
        L.const_array str_t (Array.of_list reversed_v)
      | SCall("print", [e]) ->
        let expr1 = expr builder m_locals e in
          (match (type_of_lval expr1) with
            A.String -> L.build_call printf_func [| str_format_str; (expr1) |] "printf" builder
            | A.Int -> L.build_call printf_func [| int_format_str; (expr1) |] "printf" builder
            | A.Float -> L.build_call printf_func [| float_format_str; (expr1) |] "printf" builder
            | A.Bool -> 
              let bstring = L.build_select expr1 true_str false_str "select" builder in
               L.build_call printf_func [| bstring ; (expr1)|] "printf" builder
            | A.Char -> L.build_call printf_func [| char_format_str; (expr1)|] "printf" builder
            | A.Cry -> raise (Failure ("error: this type is unsupported"))
            | _ -> raise (Failure ("error: this type is unsupported"))
          )
          | SCall(f, args) ->
            let (fdef, fdecl) = StringMap.find f function_decls in
            let llargs = List.rev( List.map (expr builder m_locals) (List.rev args)) in
            let result = (match fdecl.styp with
                         A.Cry -> ""
                         | _ -> f ^ "_result") in
            L.build_call fdef (Array.of_list llargs) result builder
         
          in
        let add_terminal (_, builder) instr =
          match L.block_terminator (L.insertion_block builder) with
            Some _ -> ()
            | None -> ignore (instr builder) in

        let rec stmt (m_locals, builder) = function
          | SBlock sl -> List.fold_left stmt (m_locals, builder) sl
          | SExpr e -> ignore(expr builder m_locals e); (m_locals, builder)
          | SIf (predicate, then_stmt, else_stmt) ->
            let bool_val = expr builder m_locals predicate in
              let merge_bb = L.append_block context "merge" the_function in
            let build_br_merge = L.build_br merge_bb in
              let then_bb = L.append_block context "then" the_function in
                add_terminal (stmt (m_locals, (L.builder_at_end context then_bb)) then_stmt)
                build_br_merge;

	            let else_bb = L.append_block context "else" the_function in
	            add_terminal (stmt (m_locals, (L.builder_at_end context else_bb)) else_stmt)
                build_br_merge;
            ignore(L.build_cond_br bool_val then_bb else_bb builder);
            m_locals, (L.builder_at_end context merge_bb)

          | SWhile (predicate, body) ->
            let pred_bb = L.append_block context "while" the_function in
            ignore(L.build_br pred_bb builder);

            let body_bb = L.append_block context "while_body" the_function in
	            add_terminal (stmt (m_locals, (L.builder_at_end context body_bb)) body)
	            (L.build_br pred_bb);

	          let pred_builder = L.builder_at_end context pred_bb in
	          let bool_val = expr pred_builder m_locals predicate in

	          let merge_bb = L.append_block context "merge" the_function in
	          ignore(L.build_cond_br bool_val body_bb merge_bb pred_builder);
	          m_locals,  (L.builder_at_end context merge_bb)

          | SFor (e1, e2, e3, body) -> stmt (m_locals, builder)
          ( SBlock [SExpr e1 ; SWhile (e2, SBlock [body ; SExpr e3]) ] )
          | SReturn e -> ignore(match fdecl.styp with
                              (* Special "return nothing" instr *)
                              A.Cry -> L.build_ret_void builder
                              (* Build return statement *)
                            | _ -> L.build_ret (expr builder m_locals e) builder );
                    m_locals, builder
          | SVdecl (t, v) -> let local = L.build_alloca (ltype_of_typ t) v builder in
          let m_locals = StringMap.add v local m_locals in (m_locals, builder)
        |SVdeclAssign (typ, v, ((t, e): sexpr)) ->
          let e' =  expr builder m_locals (t, e) in
        let add_local m (t, n) =
          let local_var = L.build_alloca (ltype_of_typ t) n builder
          in StringMap.add n local_var m
        in
        let m_locals = add_local m_locals (typ, v)
        in ignore(L.build_store e' (lookup v m_locals) builder); (m_locals, builder)
      in
      let (_, builder) = stmt (locals, builder) (SBlock fdecl.sbody) in
      add_terminal (m_locals, builder) (match fdecl.styp with
        A.Cry -> L.build_ret_void
        | A.Int -> L.build_ret (L.const_int i32_t 0)
        | A.Float -> L.build_ret (L.const_float double_t 0.0)
        | A.Char -> L.build_ret (L.const_int i8_t (Char.code '0'))
        | A.Bool -> L.build_ret (L.const_int i1_t 0) 
        | t -> L.build_ret (L.const_int (ltype_of_typ t) 0))

    in List.iter (build_function_body StringMap.empty) functions;
    the_module
